/* FILE: mmdispcmds.cc                 -*-Mode: c++-*-
 *
 * Tcl commands implemented in C/C++ and supported by shell mmdispsh
 *
 */

#include <math.h>	/* fmod(), ... */
#include <stdlib.h>	/* atoi(), ... */
#include <string.h>	/* strncpy(), ... */

#include "oc.h"
#include "nb.h"
#include "vf.h"

#include "mmdispcmds.h"
#include "display.h"	/* DisplayFrame, CoordinateSystem */
#include "bitmap.h"

/* End includes */

static Vf_Mesh *myMesh;
static DisplayFrame myFrame;

#define MY_BUF_SIZE 4096
static CHAR buf[MY_BUF_SIZE];

static REAL4m CoordsToAngle(CoordinateSystem coords)
{
  REAL4m rotang=0.;
  switch(coords) {
  case CS_CALCULATION_STANDARD: // Safety
  case CS_DISPLAY_STANDARD:  rotang=0.;   break;
  case CS_DISPLAY_ROT_90:    rotang=90.;  break;
  case CS_DISPLAY_ROT_180:   rotang=180.; break;
  case CS_DISPLAY_ROT_270:   rotang=270.; break;
  case CS_ILLEGAL:
  default:
    PlainError(1,"Unrecognized or illegal coordinate system (%d)"
	       " detected in CoordsToAngle() (File tkcmds.cc)\n",
	       int(coords));
    break; // Shouldn't get to here
  }
  return rotang;
}

static CoordinateSystem AngleToCoords(REAL4m angle_in_degrees)
{ // Rounds to nearest supported coordinate system rotation
  CoordinateSystem coords=CS_ILLEGAL;
  int quad=int(OC_ROUND(fmod(double(angle_in_degrees),
			     double(360.))/90.));
  if(quad<0) quad+=4;  if(quad>3) quad-=4;
  switch(quad) {
  case 0:  coords=CS_DISPLAY_STANDARD;  break;
  case 1:  coords=CS_DISPLAY_ROT_90;    break;
  case 2:  coords=CS_DISPLAY_ROT_180;   break;
  case 3:  coords=CS_DISPLAY_ROT_270;   break;
  default:
    PlainError(1,"Programming error: Illegal coordinate system (quad=%d)"
	       " occurred in AngleToCoords() (File tkcmds.cc)\n",quad);
    break; // Shouldn't get to here
  }
  return coords;
}

int GetMeshType(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshType must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetMeshType();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetVecColor(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetVecColor must be called with 1 argument:"
	    " A 3 element list representing a vector"
            " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int carr_size;
  char** carr;
  if (Tcl_SplitList(interp,argv[1],&carr_size,&carr)!=TCL_OK) {
    return TCL_ERROR;
  }
  if(carr_size!=3) {
    sprintf(buf,"Input list has %d != 3 elements",carr_size);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  Nb_Vec3<REAL4> v(Nb_Atof(carr[0]),Nb_Atof(carr[1]),Nb_Atof(carr[2]));
  Tcl_Free((char *)carr);
  myFrame.GetVecColor(v);
  Oc_AutoBuf ab;
  Tcl_AppendResult(interp,ab(myFrame.GetVecColor(v)),(char *)NULL);
  return TCL_OK;
}


int GetMeshName(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshName must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetName();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshTitle(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshTitle must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *title=myMesh->GetTitle();
  if(title!=NULL) strncpy(buf,title,sizeof(buf));
  else            buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int SetMeshTitle(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetMeshTitle must be called with 1 argument:"
	    " new_title (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  myMesh->SetTitle(argv[1]);
  return TCL_OK;
}

int GetMeshDescription(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshDesc must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *desc=myMesh->GetDescription();
  if(desc!=NULL) strncpy(buf,desc,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshIncrement(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshIncrement must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  sprintf(buf,"%f",myMesh->GetSubsampleGrit());
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int ChangeMesh(ClientData cd,Tcl_Interp *interp,int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<5 || argc>6) {
    sprintf(buf,"ChangeMesh must be called with 4 or 5 arguments:"
	    " mesh_filename"
	    " frame_width frame_height rotation_degrees"
	    " [zoom]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  delete myMesh;
  if(argv[1]==NULL || argv[1][0]=='\0') { // Null request
    myMesh=new Vf_EmptyMesh();
  }
  else {
    Vf_FileInput *vffreader
      = Vf_FileInput::NewReader(argv[1]);
    if(vffreader==NULL) { // Unknown file type
      myMesh=new Vf_EmptyMesh();
    }
    else { // Reader found for file argv[1]
      myMesh=vffreader->NewMesh();
      delete vffreader;
      if(myMesh==NULL) myMesh=new Vf_EmptyMesh(); // Safety
    }
  }

  myFrame.SetMesh(myMesh);
  REAL4m width=Nb_Atof(argv[2]);
  REAL4m height=Nb_Atof(argv[3]);
  CoordinateSystem coords=AngleToCoords(Nb_Atof(argv[4]));
  REAL4m zoom=-1.;
  if(argc>5) {
    zoom=Nb_Atof(argv[5]);
  }

  // Adjust sizes to leave room for margins.
  Oc_AutoBuf ab1,ab2;
  char *cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("misc,margin"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0') {
    REAL4m margin=Nb_Atof(cptr);
    width-=2*margin;
    height-=2*margin;
  }

  myFrame.SetCoordinates(coords);

  if(zoom>0.) myFrame.SetZoom(zoom);
  else        myFrame.SetZoom(width,height);

  UpdatePlotConfiguration(cd,interp,0,NULL);
  /// For now, just use current configuration /**/

  return TCL_OK;
}

int DrawFrame(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  /*
   * We use the following buffer to store commands for passing to
   * Tcl_Eval().  It is important that we do not use the global
   * buffer, buf, for commands evaluted by Tcl_Eval() since that is
   * used to write result strings to Tcl_AppendResult().  If the
   * contents of a buffer are overwritten while Tcl_Eval() is
   * evaluating them, bad results can happen.
   *
   * Note that similar bad things can happen if the contents of
   * cmdbuf are overwritten which Tcl_Eval() evaluates them.  This
   * means it is a bad thing for DrawFrame() to appear on the call
   * stack twice.  Take care that none of the Tcl_Eval()s below make
   * another call to DrawFrame().
   */
  static CHAR cmdbuf[MY_BUF_SIZE];

  int error_code;
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"DrawFrame must be called with 1 argument: canvas_name"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(cmdbuf,"%s delete all ;# update idletasks",argv[1]);
  /// Uncomment the "update idletasks" command to get a white
  /// screen between frames
  Tcl_Eval(interp,cmdbuf); // Clear canvas
  error_code=myFrame.Show(argv[1]); // Redraw canvas

  sprintf(cmdbuf,"global DrawFrameCount; incr DrawFrameCount");
  Tcl_Eval(interp,cmdbuf); // Increment DrawFrameCount variable

/* Adjust subsample setting to actual value.
 * Note: Any tracings on plot_config triggered as a result
 *  of the Tcl_SetVar2 calls will send the name used in the
 *  call (here, e.g., "plot_config arrow_subsample") to the
 *  trace callback, and the call will appear to be from the
 *  proc that called this routine (DrawFrame).  If that routine
 *  has not made plot_config visible through a 'global plot_config'
 *  command, then a "upvar plot_config" will *fail* in the trace
 *  callback.  To protect against this, trace callbacks on plot_config
 *  should have the global name hardcoded into the trace call, and the
 *  callback itself should use the hardcoded value instead of the
 *  "name" argument.  For example:
 *
 *            trace variable plot_config(arrow_subsample) w \
 *                "MyCallback plot_config"
 *  and
 *        proc MyCallback { globalname name elt op } {
 *                upvar #0 $globalname x
 *                ...
 *        }
 *
 *  (The "elt" value can be either hardcoded into globalname, or
 *  processed separately.)
 */
  Oc_AutoBuf ab1,ab2;
  REAL4m arrow_sample,pixel_sample;
  REAL4m arrow_sample_request,pixel_sample_request;
  myFrame.GetRequestedSubsampleRates(arrow_sample_request,
				     pixel_sample_request);
  myFrame.GetActualSubsampleRates(arrow_sample,pixel_sample);
  if(arrow_sample_request<0.) {
    arrow_sample*=-1;
    if(arrow_sample > -0.01) arrow_sample = -0.01;
  }
  sprintf(buf,"%g",arrow_sample);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("arrow,subsample"),
	      buf,TCL_GLOBAL_ONLY);
  if(pixel_sample_request<0.) {
    pixel_sample*=-1;
    if(pixel_sample > -0.01) pixel_sample = -0.01;
  }
  sprintf(buf,"%g",pixel_sample);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("pixel,subsample"),
	      buf,TCL_GLOBAL_ONLY);

  if(error_code!=TCL_OK) {
    sprintf(buf,"Error from myFrame.Show(%s)",argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return error_code;
  }
  return TCL_OK;
}

int GetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetFrameRotation must be called with"
	    " 1 argument: name of Tk's Frame Rotation variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m rotang=int(OC_ROUND(CoordsToAngle(myFrame.GetCoordinates())));

  sprintf(buf,"%d",rotang);
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

// Following routine added by SLW May-1998, modified by MJD June-1998.
// It copies the plot configuration values from the Tcl variable
// plot_config into the corresponding Frame plot_config structures.
// Also sets quantitylist value from Frame.
int UpdatePlotConfiguration(ClientData,Tcl_Interp *interp,int,char **)
{
  Tcl_ResetResult(interp);
  // Ignore argument list; just read plot_config() directly and update
  // all fields.
  char *cptr; int colorcount;
  Oc_AutoBuf ab1,ab2;
  PlotConfiguration arrow_config,pixel_config;
  myFrame.GetPlotConfiguration(arrow_config,pixel_config);

  // Arrow plot configuration
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,status"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.displaystate=atoi(cptr);

  colorcount=-1;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("arrow,colorcount"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0') colorcount=atoi(cptr);
  if(colorcount<0) colorcount=arrow_config.colormap.GetColorCount();
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("arrow,colormap"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.colormap.Setup(colorcount,cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,quantity"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.colorquantity=cptr;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,subsample"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.subsample=Nb_Atof(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("arrow,size"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    arrow_config.size=Nb_Atof(cptr);



  // Pixel plot configuration
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,status"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.displaystate=atoi(cptr);

  colorcount=-1;
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
		   ab2("pixel,colorcount"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0') colorcount=atoi(cptr);
  if(colorcount<0) colorcount=pixel_config.colormap.GetColorCount();
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),
			 ab2("pixel,colormap"),TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.colormap.Setup(colorcount,cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,quantity"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.colorquantity=cptr;

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,subsample"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.subsample=Nb_Atof(cptr);

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("pixel,size"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    pixel_config.size=Nb_Atof(cptr);

  // Misc.
  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("misc,drawboundary"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    myFrame.SetDrawBoundary(atoi(cptr));

  cptr=Tcl_GetVar2(interp,ab1("plot_config"),ab2("misc,background"),
		   TCL_GLOBAL_ONLY);
  if(cptr!=NULL && *cptr!='\0')
    myFrame.SetBackgroundColor(cptr);

  // Get and set quantity type list
  Nb_List<Nb_DString> cqtypes;
  int cqtypec=myFrame.GetColorQuantityTypes(cqtypes);
  char **cqtypev=new char*[cqtypec];
  int i; Nb_DString* sp;
  for(i=0, sp=cqtypes.GetFirst() ;
      sp!=NULL && i<cqtypec ; i++,sp=cqtypes.GetNext()) {
    cqtypev[i]=new char[sp->Length()+1];
    strcpy(cqtypev[i],sp->GetStr());
  }
  int itop=i;
  char *cqlist=Tcl_Merge(cqtypec,cqtypev);
  Tcl_SetVar2(interp,ab1("plot_config"),ab2("quantitylist"),
			 cqlist,TCL_GLOBAL_ONLY);
  Tcl_Free(cqlist);
  for(i=0;i<itop;i++) delete[] cqtypev[i];
  delete[] cqtypev;

  // Try to make selected color quantities valid
  int acq_match=0,pcq_match=0;
  const char* dfcq=NULL;
  for(sp=cqtypes.GetFirst() ; sp!=NULL ; sp=cqtypes.GetNext()) {
    if(strcmp(arrow_config.colorquantity,sp->GetStr())==0) acq_match=1;
    if(strcmp(pixel_config.colorquantity,sp->GetStr())==0) pcq_match=1;
    if(dfcq==NULL) dfcq=sp->GetStr();
  }
  if(!acq_match && dfcq!=NULL) {
    arrow_config.colorquantity=dfcq;
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("arrow,quantity"),
		Oc_AutoBuf(dfcq),TCL_GLOBAL_ONLY);
  }
  if(!pcq_match && dfcq!=NULL) {
    pixel_config.colorquantity=dfcq;
    Tcl_SetVar2(interp,ab1("plot_config"),ab2("pixel,quantity"),
		Oc_AutoBuf(dfcq),TCL_GLOBAL_ONLY);
  }

  // Apply new configuration
  myFrame.SetPlotConfiguration(arrow_config,pixel_config);
  return TCL_OK;
}

int SetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetFrameRotation must be called with"
	    " 1 argument: new_rotation_angle"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  myFrame.SetCoordinates(AngleToCoords(Nb_Atof(argv[1])));

  return TCL_OK;
}

int GetFrameBox(ClientData,Tcl_Interp *interp,
		 int argc,char**)
{ // NOTE: The return value from this routine is a *6*-tuple,
  //  "xmin ymin zmin xmax ymax zmax".
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetFrameBox should be called with no arguments."
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Nb_BoundingBox<REAL4> bb=myFrame.GetDisplayBox();
  Nb_Vec3<REAL4> minpt,maxpt;
  bb.GetExtremes(minpt,maxpt);
  if(bb.IsEmpty()) maxpt=minpt;  // Empty boxes may have maxpt<minpt.

  sprintf(buf,"%g %g %g %g %g %g",
	  minpt.x,minpt.y,minpt.z,maxpt.x,maxpt.y,maxpt.z);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetZoom(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetZoom must be called with"
	    " 1 argument: name of zoom variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%.6g",myFrame.GetZoom());
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

int SetZoom(ClientData,Tcl_Interp *interp,
		  int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<2 || argc>3) {
    sprintf(buf,"SetZoom must be called with"
	    " either 1 or 2 arguments: <new_zoom|new_width new_height>"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Change mesh scale in display list
  REAL4m newzoom;
  if(argc==2) newzoom=myFrame.SetZoom(Nb_Atof(argv[1]));
  else        newzoom=myFrame.SetZoom(Nb_Atof(argv[1]),Nb_Atof(argv[2]));

  sprintf(buf,"%g",newzoom);
  Tcl_SetResult(interp,buf,TCL_VOLATILE);

  return TCL_OK;
}

int GetDataValueUnit(ClientData,Tcl_Interp *interp,
		     int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetDataValueUnit must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%.*s",MY_BUF_SIZE,myFrame.GetValueUnit());
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}

int GetDefaultDataValueScaling(ClientData,Tcl_Interp *interp,
			       int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetDefaultDataValueScaling must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%.6g",myFrame.GetDefaultValueScaling());
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}

int GetDataValueScaling(ClientData,Tcl_Interp *interp,
			int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetDataValueScaling must be called with"
	    " no arguments (%d arguments passed).",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%.6g",myFrame.GetValueScaling());
  Tcl_AppendResult(interp,buf,(char *)NULL);

  return TCL_OK;
}


int SetDataValueScaling(ClientData,Tcl_Interp *interp,
		      int argc,char** argv)
{
  // Adjust data value scaling
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetDataValueScale must be called with"
	    " 1 arguments: <new_scale>"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL8m newscale=myFrame.SetValueScaling(Nb_Atof(argv[1]));
  sprintf(buf,"%g",newscale);
  Tcl_SetResult(interp,buf,TCL_VOLATILE);

  return TCL_OK;
}

int WriteMeshUsingDeprecatedVIOFormat(ClientData,Tcl_Interp *interp,
	      int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc != 2) {
    sprintf(buf,"WriteMesh must be called with"
	    " 1 argument: filename"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int errcode=TCL_OK;

  // Open file
  FILE *fptr;
  if (argv[1][0] == '\0') {
    fptr=stdout;
  } else {
    fptr=Nb_FOpen(argv[1],"wb");
  }
  if(fptr==NULL) {
    sprintf(buf,"Error in %s: Unable to open file %s for output",
	    argv[0], argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    errcode=TCL_ERROR;
  }
  else {
    Vf_VioFileOutput vfo;
    if(vfo.WriteMesh(myMesh,fptr,NULL)!=0) {
      sprintf(buf,"WriteMesh error");
      Tcl_AppendResult(interp,buf,(char *)NULL);
      errcode=TCL_ERROR;
    }
    fflush(fptr);
    if (argv[1][0] != '\0') {
      fclose(fptr);
    }
  }

  return errcode;
}

int WriteMesh(ClientData,Tcl_Interp *interp,
	      int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<4 || argc>6) {
    sprintf(buf,"WriteMesh must be called with"
	    " 3-5 arguments: filename <text|binary4|binary8>"
	    " <regular|irregular> [title] [description]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int errcode=TCL_OK;

  Vf_OvfDataStyle ods=vf_oascii;
  if(strcmp("binary4",argv[2])==0)      ods=vf_obin4;
  else if(strcmp("binary8",argv[2])==0) ods=vf_obin8;

  BOOL force_irreg=0;
  if(strcmp("irregular",argv[3])==0) force_irreg=1;

  const char *title="",*desc="";
  if(argc>4) title=argv[4];
  if(argc>5) desc=argv[5];

  // Open file
  FILE *fptr;
  if (argv[1][0] == '\0') {
    fptr=stdout;
  } else {
    fptr=Nb_FOpen(argv[1],"wb");
  }
  if(fptr==NULL) {
    sprintf(buf,"Error in %s: Unable to open file %s for output", 
	    argv[0], argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    errcode=TCL_ERROR;
  }
  else {
    Vf_OvfFileOutput ofo;
    if(ofo.WriteMesh(myMesh,fptr,ods,force_irreg,title,desc)!=0) {
      sprintf(buf,"WriteMesh error");
      Tcl_AppendResult(interp,buf,(char *)NULL);
      errcode=TCL_ERROR;
    }
    fflush(fptr);
    if (argv[1][0] != '\0') {
      fclose(fptr);
    }
  }

  return errcode;
}

int
WriteMeshAverages(ClientData, Tcl_Interp *interp, int argc, char** argv)
{
    Tcl_Channel channel;
    int mode;

    Tcl_ResetResult(interp);
    if(argc!=2) {
      sprintf(buf,"WriteMeshAverages must be called with 1 argument:"
  	    " an output channel (%d arguments passed)",argc-1);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    if ((channel = Tcl_GetChannel(interp, argv[1], &mode)) == NULL) {
      return TCL_ERROR;
    }
    if (!(mode & TCL_WRITABLE)) {
      sprintf(buf,"%s is not a writable channel", argv[1]);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }

    // Is this a rectangular mesh?
    if(strcmp(myMesh->GetMeshType(),"Vf_GridVec3f")!=0) {
      sprintf(buf,"Invalid mesh type: %s\n",myMesh->GetMeshType());
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }

    // Downcast
    Vf_GridVec3f& grid = *((Vf_GridVec3f *)myMesh);

    // Collect and print x-data
    int i,j,k,isize,jsize,ksize;
    grid.GetDimens(isize,jsize,ksize);
    if(isize<1 || jsize<1 || ksize<1) {
      sprintf(buf,"Bad mesh dimensions: %d %d %d",isize,jsize,ksize);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    Nb_Vec3<REAL8> *m=new Nb_Vec3<REAL8>[isize];
    for(i=0;i<isize;i++) m[i].Set(0.,0.,0.);
    for(i=0;i<isize;i++) for(j=0;j<jsize;j++) for(k=0;k<ksize;k++) {
      m[i]+=grid(i,j,k);
    }
    Nb_Vec3<REAL8> base=grid.GetBasePoint();
    Nb_Vec3<REAL8> step=grid.GetGridStep();

    double mult=(1./double(jsize*ksize));
    sprintf(buf,"# Averages across x axis (%d points each)\n", jsize*ksize);
    Tcl_Write(channel, buf, strlen(buf));
    for(i=0;i<isize;i++) {
      m[i]*=mult;
      sprintf(buf,"%- #12.6g    %- #12.6g  %- #12.6g  %- #12.6g\n",
              base.x+double(i)*step.x,m[i].x,m[i].y,m[i].z);
      Tcl_Write(channel, buf, strlen(buf));
    }
    Tcl_Flush(channel);
    delete[] m;

    // Collect and print y-data
    m=new Nb_Vec3<REAL8>[jsize];
    for(j=0;j<jsize;j++) m[j].Set(0.,0.,0.);
    for(i=0;i<isize;i++) for(j=0;j<jsize;j++) for(k=0;k<ksize;k++) {
      m[j]+=grid(i,j,k);
    }
    mult=(1./double(isize*ksize));
    sprintf(buf,"\n\n# Averages across y axis (%d points each)\n",isize*ksize);
    Tcl_Write(channel, buf, strlen(buf));
    for(j=0;j<jsize;j++) {
      m[j]*=mult;
      sprintf(buf,"%- #12.6g    %- #12.6g  %- #12.6g  %- #12.6g\n",
              base.y+double(j)*step.y,m[j].x,m[j].y,m[j].z);
      Tcl_Write(channel, buf, strlen(buf));
    }
    Tcl_Flush(channel);
    delete[] m;

    // Collect and print z-data
    m=new Nb_Vec3<REAL8>[isize];
    for(k=0;k<ksize;k++) m[k].Set(0.,0.,0.);
    for(i=0;i<isize;i++) for(j=0;j<jsize;j++) for(k=0;k<ksize;k++) {
      m[k]+=grid(i,j,k);
    }
    mult=(1./double(isize*jsize));
    sprintf(buf,"\n\n#Averages across z axis (%d points each)\n", isize*jsize);
    Tcl_Write(channel, buf, strlen(buf));
    for(k=0;k<ksize;k++) {
      m[k]*=mult;
      sprintf(buf, "%- #12.6g    %- #12.6g  %- #12.6g  %- #12.6g\n",
              base.z+double(k)*step.z,m[k].x,m[k].y,m[k].z);
      Tcl_Write(channel, buf, strlen(buf));
    }
    Tcl_Flush(channel);
    delete[] m;

    return TCL_OK;
}

int
BitmapCmd(ClientData clientData, Tcl_Interp *interp, int argc, char** argv)
{
    char c;
    int margin, width, height, crop, isNew, mode;
    int xmin, ymin, xmax, ymax, xmarginadj, ymarginadj;
    OommfBitmap *bitmapPtr;
    Nb_BoundingBox<REAL4> bbox;
    Nb_Vec3<REAL4> minpt, maxpt;
    Oc_AutoBuf ab1, ab2;
    Tcl_HashTable *tablePtr = (Tcl_HashTable *)clientData;
    Tcl_HashEntry *entryPtr;
    Tcl_Channel channel;

    if (argc < 2) {
	sprintf(buf, "wrong # args: should be \"%s option bitmapName"
		" ?arg ...?\"", argv[0]);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    if ((c == 'C') && (strcmp(argv[1], "Create") == 0)) {
	if (argc != 3) {
	    sprintf(buf, "wrong # args: should be \"%s Create bitmapName\"",
		    argv[0]);
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	    return TCL_ERROR;
	}
	entryPtr = Tcl_CreateHashEntry(tablePtr, argv[2], &isNew);
	if (!isNew) {
	    sprintf(buf, "bitmap \"%s\" already exists", argv[2]);
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	    return TCL_ERROR;
	}

	// Create a new OommfBitmap from the global DisplayFrame
	bitmapPtr = new OommfBitmap;

	margin = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
		ab2("misc,margin"), TCL_GLOBAL_ONLY));
	width = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
		ab2("misc,width"), TCL_GLOBAL_ONLY));
	height = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
		ab2("misc,height"), TCL_GLOBAL_ONLY));
	crop = atoi(Tcl_GetVar2(interp, ab1("plot_config"),
		ab2("misc,crop"), TCL_GLOBAL_ONLY));

	bbox = myFrame.GetDisplayBox();
	bbox.GetExtremes(minpt, maxpt);
	xmin = int(floor(minpt.x)) - margin;
	ymin = int(floor(minpt.y)) - margin;
	xmax = int(ceil(maxpt.x)) + margin;
	ymax = int(ceil(maxpt.y)) + margin;
	xmarginadj = (xmax - xmin + 1) - width;
	ymarginadj = (ymax - ymin + 1) - height;
	if(xmarginadj>0) {
	    // Resulting size is bigger than requested. Trim.
	    xmin += int(floor(xmarginadj/2.));
	    xmax -= int(ceil(xmarginadj/2.));
	}
	if(ymarginadj>0) {
	    // Resulting size is bigger than requested. Trim.
	    ymin += int(floor(ymarginadj/2.));
	    ymax -= int(ceil(ymarginadj/2.));
	}
	if(!crop) {
	    // No-crop request, so make bitmap exact size requested
	    xmax = xmin + width - 1;
	    ymax = ymin + height - 1;
	}
	bitmapPtr->Setup(xmin,ymin,xmax,ymax);
	myFrame.FillBitmap(*bitmapPtr);
	Tcl_SetHashValue(entryPtr, (ClientData)bitmapPtr);
	return TCL_OK;
    }
    if ((c == 'D') && (strcmp(argv[1], "Delete") == 0)) {
	if (argc != 3) {
	    sprintf(buf, "wrong # args: should be \"%s Delete bitmapName\"",
		    argv[0]);
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	    return TCL_ERROR;
	}
	entryPtr = Tcl_FindHashEntry(tablePtr, argv[2]);
	if (entryPtr == NULL) {
	    sprintf(buf, "can't find bitmap \"%s\"", argv[2]);
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	    return TCL_ERROR;
	}
	Tcl_DeleteHashEntry(entryPtr);
	return TCL_OK;
    }
    if ((c == 'W') && (strcmp(argv[1], "Write") == 0)) {
	if (argc != 5) {
	    sprintf(buf, "wrong # args: should be "
		    "\"%s Write bitmapName format channel\"", argv[0]);
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	    return TCL_ERROR;
	}
	if ((channel = Tcl_GetChannel(interp, argv[4], &mode)) == NULL) {
	    return TCL_ERROR;
	}
	if (!(mode & TCL_WRITABLE)) {
	    sprintf(buf,"%s is not a writable channel", argv[4]);
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	    return TCL_ERROR;
	}
	entryPtr = Tcl_FindHashEntry(tablePtr, argv[2]);
	if (entryPtr == NULL) {
	    sprintf(buf, "can't find bitmap \"%s\"", argv[2]);
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	    return TCL_ERROR;
	}
	bitmapPtr = (OommfBitmap *) Tcl_GetHashValue(entryPtr);
	if ( strcmp("P3", argv[3]) == 0) {
	    bitmapPtr->WritePPMChannel(channel, 3);
	    return TCL_OK;
	}
	if ( strcmp("P6", argv[3]) == 0) {
	    bitmapPtr->WritePPMChannel(channel, 6);
	    return TCL_OK;
	}
	if ( strcmp("B24", argv[3]) == 0) {
	    bitmapPtr->WriteBMPChannel(channel, 24);
	    return TCL_OK;
	}
	sprintf(buf, "ERROR: Unknown file format request: %s\n", argv[3]);
	Tcl_AppendResult(interp, buf, (char *) NULL);
	return TCL_ERROR;
    }
    sprintf(buf,"bad option \"%s\": must be Create, Delete, or Write",argv[1]);
    Tcl_AppendResult(interp, buf, (char *) NULL);
    return TCL_ERROR;
}

void
BitmapDeleteProc(ClientData clientData)
{
    Tcl_HashTable *tablePtr = (Tcl_HashTable *) clientData;
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchId;
    OommfBitmap *bitmapPtr;

    for (entryPtr = Tcl_FirstHashEntry(tablePtr, &searchId);
	    entryPtr != NULL; entryPtr = Tcl_NextHashEntry(&searchId)) {
	bitmapPtr = (OommfBitmap *) Tcl_GetHashValue(entryPtr);
	delete bitmapPtr;
    }
    Tcl_DeleteHashTable(tablePtr);
    delete tablePtr;
}

static void
CleanupMesh(ClientData)
{
  if (myMesh!=NULL) {
    delete myMesh;
  }
}

int Mmdispcmds_Init(Tcl_Interp *interp)
{
#define RETURN_TCL_ERROR                                                \
    Tcl_AddErrorInfo(interp, ab1("\n    (in Mmdispcmds_Init())"));      \
    return TCL_ERROR

  Oc_AutoBuf ab1,ab2;
  Tcl_HashTable *bitmapTablePtr;

  // Verify run-time and compile-time versions of required support
  // extensions match.
//  if (Tcl_PkgRequire(interp, ab1("Tcl"), ab2(TCL_VERSION), 1) == NULL) {
//    Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds "), NULL);
//    Tcl_AppendResult(interp, ab1(MMDISPCMDS_VERSION), NULL);
//    Tcl_AppendResult(interp, ab1(" needs Tcl"), NULL);
//    Tcl_AppendResult(interp, ab1(TCL_VERSION), NULL);
//    Tcl_AppendResult(interp, ab1(")"), NULL);
//  }

/*
 * Need at least Oc 1.1.1.0 to get Oc_Config's compiletime_* features
 */
    if (Tcl_PkgRequire(interp, ab1("Oc"), ab2("1.1.1.0"), 0) == NULL) {
        Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds " MMDISPCMDS_VERSION
                " needs Oc 1.1.1.0)"), NULL);
        RETURN_TCL_ERROR;
    }

/*
 * Need at least Vf 1.1.1.0 to get Vf_Mesh::GetMaxMag()
 */
    if (Tcl_PkgRequire(interp, ab1("Vf"), ab2("1.1.1.0"), 0) == NULL) {
        Tcl_AppendResult(interp, ab1("\n\t(Mmdispcmds " MMDISPCMDS_VERSION
                " needs Vf 1.1.1.0)"), NULL);
        RETURN_TCL_ERROR;
    }
  
  // Record compile and run-time patch levels

  // Initialize C++ data structures
  myMesh=new Vf_EmptyMesh();
  Tcl_CreateExitHandler((Tcl_ExitProc *) CleanupMesh, 0);
  myFrame.SetTclInterp(interp);
  myFrame.SetMesh(myMesh);

  // Register C/C++ routines with interpreter
  Oc_RegisterCommand(interp,"ChangeMesh",ChangeMesh);
  Oc_RegisterCommand(interp,"DrawFrame",DrawFrame);
  Oc_RegisterCommand(interp,"GetDataValueUnit",GetDataValueUnit);
  Oc_RegisterCommand(interp,"GetDataValueScaling",GetDataValueScaling);
  Oc_RegisterCommand(interp,"GetDefaultDataValueScaling",
		     GetDefaultDataValueScaling);
  Oc_RegisterCommand(interp,"GetFrameBox",GetFrameBox);
  Oc_RegisterCommand(interp,"GetFrameRotation",GetFrameRotation);
  Oc_RegisterCommand(interp,"GetZoom",GetZoom);
  Oc_RegisterCommand(interp,"GetMeshDescription",GetMeshDescription);
  Oc_RegisterCommand(interp,"GetMeshIncrement",GetMeshIncrement);
  Oc_RegisterCommand(interp,"GetMeshName",GetMeshName);
  Oc_RegisterCommand(interp,"GetMeshTitle",GetMeshTitle);
  Oc_RegisterCommand(interp,"GetMeshType",GetMeshType);
  Oc_RegisterCommand(interp,"GetVecColor",GetVecColor);
  Oc_RegisterCommand(interp,"SetDataValueScaling",SetDataValueScaling);
  Oc_RegisterCommand(interp,"SetFrameRotation",SetFrameRotation);
  Oc_RegisterCommand(interp,"SetZoom",SetZoom);
  Oc_RegisterCommand(interp,"SetMeshTitle",SetMeshTitle);
  Oc_RegisterCommand(interp,"UpdatePlotConfiguration",
                     UpdatePlotConfiguration);
  Oc_RegisterCommand(interp,"WriteMesh",WriteMesh);
  Oc_RegisterCommand(interp,"WriteMeshUsingDeprecatedVIOFormat",
		     WriteMeshUsingDeprecatedVIOFormat);
  Oc_RegisterCommand(interp,"WriteMeshAverages",WriteMeshAverages);

  bitmapTablePtr = new Tcl_HashTable;
  Tcl_InitHashTable(bitmapTablePtr, TCL_STRING_KEYS);
  Tcl_CreateCommand(interp, ab1("Bitmap"), BitmapCmd,
      (ClientData) bitmapTablePtr, BitmapDeleteProc);

  if (Tcl_PkgProvide(interp, ab1("Mmdispcmds"), ab2(MMDISPCMDS_VERSION)) 
      != TCL_OK) {
    RETURN_TCL_ERROR;
  }
  if (Oc_InitScript(interp, "Mmdispcmds", MMDISPCMDS_VERSION) != TCL_OK) {
    RETURN_TCL_ERROR;
  }
  return TCL_OK;

#undef RETURN_TCL_ERROR

}

